"""
Initialize Online Portfolio Selection Module.
"""
# Parent Class
from famlafl.online_portfolio_selection.base import OLPS
from famlafl.online_portfolio_selection.up import UP

# Benchmarks
from famlafl.online_portfolio_selection.bah import BAH
from famlafl.online_portfolio_selection.best_stock import BestStock
from famlafl.online_portfolio_selection.crp import CRP
from famlafl.online_portfolio_selection.bcrp import BCRP

# Momentum
from famlafl.online_portfolio_selection.eg import EG
from famlafl.online_portfolio_selection.ftl import FTL
from famlafl.online_portfolio_selection.ftrl import FTRL

# Mean Reversion
from famlafl.online_portfolio_selection.rmr import RMR
from famlafl.online_portfolio_selection.pamr import PAMR
from famlafl.online_portfolio_selection.cwmr import CWMR
from famlafl.online_portfolio_selection.olmar import OLMAR

# Pattern Matching
from famlafl.online_portfolio_selection.corn import CORN
from famlafl.online_portfolio_selection.cornu import CORNU
from famlafl.online_portfolio_selection.cornk import CORNK
from famlafl.online_portfolio_selection.scorn import SCORN
from famlafl.online_portfolio_selection.scornk import SCORNK
from famlafl.online_portfolio_selection.fcorn import FCORN
from famlafl.online_portfolio_selection.fcornk import FCORNK
