"""
Initialize Benchmarks Module.
"""
from famlafl.online_portfolio_selection.bah import BAH
from famlafl.online_portfolio_selection.best_stock import BestStock
from famlafl.online_portfolio_selection.crp import CRP
from famlafl.online_portfolio_selection.bcrp import BCRP
