"""
Initialize Momentum Module.
"""
from famlafl.online_portfolio_selection.eg import EG
from famlafl.online_portfolio_selection.ftl import FTL
from famlafl.online_portfolio_selection.ftrl import FTRL
