"""
FAMLAFL helps portfolio managers and traders who want to leverage the power of machine learning by providing
reproducible, interpretable, and easy to use tools.

Adding FAMLAFL to your companies pipeline is like adding a department of PhD researchers to your team.
"""

import famlafl.util as util
import famlafl.data_structures as data_structures
import famlafl.datasets as datasets
import famlafl.multi_product as multi_product
import famlafl.filters.filters as filters
import famlafl.labeling as labeling
import famlafl.features.fracdiff as fracdiff
import famlafl.sample_weights as sample_weights
import famlafl.sampling as sampling
import famlafl.bet_sizing as bet_sizing
import famlafl.structural_breaks as structural_breaks
import famlafl.feature_importance as feature_importance
import famlafl.ensemble as ensemble
import famlafl.portfolio_optimization as portfolio_optimization
import famlafl.clustering as clustering
import famlafl.microstructural_features as microstructural_features
import famlafl.cross_validation as cross_validation
import famlafl.online_portfolio_selection as online_portfolio_selection
from famlafl.backtest_statistics import statistics
from famlafl.backtest_statistics import backtests
