"""
Initialize Pattern Matching Module.
"""
from famlafl.online_portfolio_selection.corn import CORN
from famlafl.online_portfolio_selection.cornu import CORNU
from famlafl.online_portfolio_selection.cornk import CORNK
from famlafl.online_portfolio_selection.scorn import SCORN
from famlafl.online_portfolio_selection.scornk import SCORNK
from famlafl.online_portfolio_selection.fcorn import FCORN
from famlafl.online_portfolio_selection.fcornk import FCORNK
