"""
Initialize Mean Reversion Module.
"""
from famlafl.online_portfolio_selection.rmr import RMR
from famlafl.online_portfolio_selection.pamr import PAMR
from famlafl.online_portfolio_selection.cwmr import CWMR
from famlafl.online_portfolio_selection.olmar import OLMAR
