"""
Contains the code for implementing sample weights.
"""

from famlafl.sample_weights.attribution import (get_weights_by_time_decay, get_weights_by_return,
                                                 _apply_weight_by_return)
